/**
 *  MOSAIIC Model - 9 decembre 2014
 *  Author: patricktaillandier
 *  Description: 
 */
 
model RoadTrafficComplex
 
global {   
	file shape_file_roads  <- file("../includes/roads2.shp") ;
	float max_speed <- 50 #km/#h;
	int nb_lanes <- 2 ; 
	
	file file_ssp_speed;
	geometry shape <- envelope(shape_file_roads);
	graph road_network_speed;  
	int nb_people <- 30000;
	bool mode_batch <- false;
	float car_size <- 4.0;
	
	map<road,float> general_speed_map_speed;
	
	float time_end_final <- 3 #h;
	
	
	bool use_macro <- false;
	
	int nb_arrived <- 0;
	float time_tot;
	float time_roads;
	float time_people;
	float time_sp_rcomp;
	list<road> macro_roads;
	float use_macro_factor <- 20.0;
	
	map<geometry,float> ec;
	
	init {  
		if (use_macro_factor = 0.0) {use_macro <-false;}
		create road from: shape_file_roads with:[name::string(read("name")),highway::string(get("highway")),junction::string(read("junction")),lanes::nb_lanes, maxspeed::max_speed, oneway::string(read("oneway")) replace("'",""), lanes_forward ::int (get( "lanesforwa")), lanes_backward :: int (get("lanesbackw"))] {
			create road with: [shape::line(reverse(shape.points)), maxspeed::maxspeed, lanes::lanes, linked_road::self] {
				myself.linked_road <- self;
			}
		}
		ask road {
			
			if lanes = 0 {lanes <- nb_lanes;}
			if maxspeed = 0 {maxspeed <- max_speed;}
			//capacite_max <- 1+ int(lanes * shape.perimeter/(4.0));
			
			//do manage_oneway_attribute;
		}	
		
		list<point> points <- [];
		ask road {
			point fp <- first(shape.points);
			point lp <- last(shape.points);
			if not(fp in points) {create node_ with:[location::fp];points << fp;}
			if not(lp in points) {create node_ with:[location::lp];points << lp;}
			section_size <- maxspeed;
			
		}
		road_network_speed <-  (as_driving_graph(road, node_));
	
		ask road where not (node_(each.target_node).is_traffic_signal) {
			list<agent> nrs <- node_(target_node).roads_out where (each != linked_road);
			if (length(nrs) = 1) {next_road <- road(first(nrs)); }	
		}
	
		list<string> roads_importance <- ["service","residential", "unclassified",  "tertiary_link", "tertiary",  "secondary_link", "secondary", "primary_link", "primary", "trunk_link", "trunk", "motorway_link", "motorway"];
		ask node_ {
			roads_in <- remove_duplicates(roads_in);
			roads_out <- remove_duplicates(roads_out);
			if (length(roads_in) > 1) {
				priority_roads <- roads_in where (road(each).junction = "roundabout");
				if (empty(priority_roads)) {
					int max_importance <- roads_in max_of (roads_importance index_of road(each).highway);
					priority_roads <- roads_in where ((roads_importance index_of road(each).highway) = max_importance);
				}
			}
			 
		}
		
	
		if (use_macro) {
			road_network_speed <- road_network_speed with_optimizer_type "Floyd Warshall";
	
			ec <- map<road, float>(edge_betweenness(road_network_speed));
			float val <- (ec.values sort_by (-each))[min ([length(ec) - 1, round(length(ec)/use_macro_factor)])];
			loop e over: ec.keys { 
				road(e).is_macro <-  (road(e).shape.perimeter >= (road(e).section_size * 2)) and (ec[e] > val);
			}
			do create_sections;
		}
		macro_roads <- road where each.is_macro;
		
		write "roads built";
		
		loop times: nb_people{do create_people;}
		loop times: 20 {
			list<people> moving_people <- people where (each.active and each.current_path != nil and each.final_target != nil );
			moving_people <- moving_people sort_by (- 1000000 * each.segment_index_on_road + each.distance_to_goal);
			ask moving_people {
				do driving;
			}
		}
		if (use_macro) {
			ask macro_roads {
				do capture_car;
			} 	
		}
		write "people created";
			
		
	}
	
	
	action create_people {
		create people  { 
			real_speed <- 130 #km /#h ;
			vehicle_length <- car_size;
			right_side_driving <- true;
			proba_lane_change_up <-rnd(0.5,1.0);
			proba_lane_change_down <- rnd(0.5,1.0);
			security_distance_coeff <- rnd(1.5,3.0);  
			proba_respect_priorities <- 1.0;
			proba_respect_stops <- [1.0];
			proba_block_node <- 0.0;
			proba_use_linked_road <- 0.0;
			max_acceleration <- rnd(10.0,15.0) #km/#h;
			max_speed <- 150 #km /#h ;
			speed_coeff <- rnd(0.8,1.1);

			
			do choose_a_path;
		}
	}

	action create_sections {
		ask road where (each.is_macro) {
			int nb <- max([1,round (shape.perimeter /section_size)]);
			float sec_length <- shape.perimeter / nb;
			geometry tmp_shape <- copy(shape);
			section prev <- nil;
			list<point> pts <- shape points_on (sec_length - 0.1); 
			loop i from: 0 to: nb -1 {
				geometry t_g;
				if (i +1< nb) {
					list<geometry> sg <- tmp_shape split_at pts[i+1];
					tmp_shape <- sg[1];
					t_g <- sg[0];
				} else {t_g <- tmp_shape;}
				
				create section with:[location::location]{
					shape <- t_g;
					max_speed <- myself.maxspeed #km/#h;
						color <-  #red;
					previous <- prev;
					if (previous != nil) {previous.next <- self;}
					prev <- self;
					Kmax <- myself.lanes * sec_length /  (car_size ) ;
				
					critical_concentration <- Kmax / 2;
					current_concentration <- 0.0 ;
					max_flow <- critical_concentration * max_speed ; 
					myself.sections << self;
					my_road <- myself;
				}	
			}
		} 
	}
	
	

	reflex dynamic {
		float tt <- machine_time;
		
		list<people> moving_people <- people where (each.active and each.current_path != nil and each.final_target != nil );
		moving_people <- moving_people sort_by (- 1000000 * each.segment_index_on_road + each.distance_to_goal);
		
		ask moving_people {
			do driving;
		}
		
		time_people <- time_people + machine_time - tt; 
		tt <- machine_time;
		if (use_macro) {
			ask macro_roads {
				do capture_car;
			} 
			ask macro_roads where not empty(each.cars){
				do macro_behavior;
			}
		}
		time_roads <- time_roads + machine_time - tt;
		
		
		time_tot <- time_tot + time_people + time_roads ;
	}
	
	reflex end_sim when: empty(people) {
		save ("cycle: " + cycle +" use_macro: " + use_macro +  " -> " + " time_people: " + time_people /1000 + " time_roads:"+ time_roads/1000 + " time_sp_rcomp: " + 
			time_sp_rcomp/1000 + " time tot: " + (time_people + time_roads)/1000 + " people: " + (people count each.active) + " cars: " +  
			sum(road collect length(each.cars)) + " total: " + length(people) + " nb road macro: " + road count each.is_macro
		) to: "result_" + use_macro + "_" + nb_people+ "_" + use_macro_factor +"_"+ seed+ ".txt" rewrite: false;
		
		write "cycle: " + cycle +" use_macro: " + use_macro +  " use_macro_factor: " + use_macro_factor +" -> " + (time_tot)/1000 + " time_people: " + time_people /1000 + " time_roads:"+ time_roads/1000  + " time_sp_rcomp: " + time_sp_rcomp/1000 + " time tot: " + (time_people + time_roads)/1000 + " people: " + (people count each.active) + " cars: " +  sum(road collect length(each.cars)) + " total: " + length(people) + " nb road macro: " + road count each.is_macro;
	
		if (not mode_batch)	{	do pause;}
		else {
			road_network_speed <- road_network_speed with_weights (road as_map (each::1.0));
		}
	}
	reflex info_time when: (cycle mod 600) = 0{
			save ("cycle: " + cycle +" use_macro: " + use_macro +  " -> " + " time_people: " + time_people /1000 + " time_roads:"+ time_roads/1000 + " time_sp_rcomp: " + 
			time_sp_rcomp/1000 + " time tot: " + (time_people + time_roads)/1000 + " people: " + (people count each.active) + " cars: " +  
			sum(road collect length(each.cars)) + " total: " + length(people) + " nb road macro: " + road count each.is_macro
		) to: "result_" + use_macro + "_" + nb_people+ "_" + use_macro_factor + ".txt" rewrite: false;
			write "cycle: " + cycle +" use_macro: " + use_macro +  " use_macro_factor: " + use_macro_factor +"_"+ seed+" -> " + " time_people: " + time_people /1000 + " time_roads:"+ time_roads/1000 + " time_sp_rcomp: " + time_sp_rcomp/1000 + " time tot: " + (time_people + time_roads)/1000 + " people: " + (people count each.active) + " cars: " +  sum(road collect length(each.cars)) + " total: " + length(people) + " nb road macro: " + road count each.is_macro;
	}
} 


	
species node_ skills: [skill_road_node] frequency: 0 {
	bool is_traffic_signal;
	int counter ;
	rgb color_fire;
	bool is_crossing;
	list<road> ways1;
	list<road> ways2;
	bool is_green;
	int time_to_change <- 60;
	list<road> neighbours_tj;
	int id;
	float weight <- 1.0;
	
	action compute_crossing{
		if (is_crossing and not(empty(roads_in))) or (length(roads_in) >= 2) {
			road rd0 <- road(roads_in[0]);
			list<point> pts <- rd0.shape.points;						
			float ref_angle <-  float( last(pts) direction_to rd0.location);
			loop rd over: roads_in {
				list<point> pts2 <- road(rd).shape.points;						
				float angle_dest <-  float( last(pts2) direction_to rd.location);
				float ang <- abs(angle_dest - ref_angle);
				if (ang > 45 and ang < 135) or  (ang > 225 and ang < 315) {
					add road(rd) to: ways2;
				}
			}
		}
	}
	
	action to_green {
		stop[0] <-  ways2 ;
		color_fire <- rgb("green");
		is_green <- true;
	}
	
	action to_red {
		stop[0] <- ways1;
		color_fire <- rgb("red");
		is_green <- false;
	}
	action dynamic_node  {
		counter <- counter + 1;
		
		if (counter >= time_to_change) { 
			counter <- 0;
			if is_green {do to_red;}
			else {do to_green;}
		} 
	}
	
	aspect base {    
		draw square(20) color:  #blue border: #black;
		
	} 
	
}

species road skills: [skill_road] frequency: 0 { 
	string oneway;
	geometry geom_display;
	int nb_people_tot <- 0;
	string type; 
	int lanes_backward;
	int lanes_forward;
	
	string junction;
	int capacite_max ;
	
	string highway;
	list<node_> neighbours;
	
	list<section> sections;
	list<people> cars;
	bool is_macro <- false;
	float section_size;
	road next_road;

	aspect geom {    
		draw geom_display border:  rgb("gray")  color: rgb("gray") ;
	}  
	
	aspect base {    
		draw shape color: rgb("black") ;
	} 
	
	aspect carto {
		if (is_macro) {draw (shape + 5) color: #red ;} 
		else {
			if type = "motorway" {draw (shape + 3) color: #orange border: #black; }
			else if type = "main_road" {draw (shape + 2) color: #yellow border: #black; }
			else {draw shape color: #black;}
			
		}
	}
	
	aspect base_gray {    
		draw shape color: rgb("gray") ;
	} 
	
	
	action manage_oneway_attribute {
		if (lanes_forward > 0 and lanes_backward > 0) {
			create road {
				lanes <-myself.lanes_backward;
				highway <- myself.highway;
				oneway <- myself.oneway;
				shape <- polyline(reverse(myself.shape.points));
				capacite_max <- 1+ int(lanes * shape.perimeter/(4.0));
				maxspeed <- myself.maxspeed;
				geom_display  <- myself.geom_display;
				linked_road <- myself;
				myself.linked_road <- self;
			}
			lanes <- lanes_forward;
			capacite_max <- 1+ int(lanes * shape.perimeter/(4.0));
		} else {
			switch oneway  {
				match "yes" ;
				match "-1" {
					shape <- polyline(reverse(shape.points));
				} 
				default {
					if (junction != 'roundabout' and junction != "'roundabout'") {
						create road {
							lanes <- max([1, int (myself.lanes / 2.0)]);
							shape <- polyline(reverse(myself.shape.points));
							capacite_max <- 1+ int(lanes * shape.perimeter/(4.0));
							maxspeed <- myself.maxspeed;
							geom_display  <- myself.geom_display;
							linked_road <- myself;
							myself.linked_road <- self;
						}
						lanes <- int(lanes /2.0 + 0.5);	
						capacite_max <- 1+ int(lanes * shape.perimeter/(4.0));
					}
				}
			}
		}
		
	}
	
	
	action capture_car  {
		if (not empty(all_agents)) {
			list<people> drivers <- list<people>(all_agents) where (each.distance_to_goal > car_size);
			if (not empty(drivers)) {
				section fs <- first(sections);
				loop dr over: drivers {do unregister(dr);}
				
				cars <- cars + drivers;
				ask drivers {active <- false;}
				ask fs {
					current_concentration <- current_concentration + length(drivers);
				}	
			}
		}
		
		
	}
	action macro_behavior {
		ask sections {
			 do compute_demand_offer;
			}
		ask sections  {
				float next_offer <- 0.0;
				if (next = nil) {
					next_offer <- self.demand;
					float val  <- step/myself.section_size * demand;
					int new_agents <- int(val);
					remaining_car <- remaining_car + (val - new_agents);
					int nbcars <- length(my_road.cars);
					if (remaining_car  >= 1.0 and new_agents < nbcars) {
						remaining_car <- remaining_car - 1;
						new_agents <- new_agents + 1;
					} else if (nbcars = 1 and remaining_car > 0.9) {
						remaining_car <- 0.0;
						new_agents <- new_agents + 1;
					}
					if (new_agents > 0) {
						list<people> cars_to_release <- new_agents first(my_road.cars);
						ask cars_to_release {
							location <- last(myself.shape.points) ;
							if ((location distance_to target_node.location) < 5.0){
								myself.my_road.cars >> self;
								myself.current_concentration <- myself.current_concentration - 1;
								do arrive;
							}
						}
						cars_to_release <- cars_to_release where not dead(each);
						ask my_road {
							if (next_road != nil and next_road.is_macro) {
									section fs <- first(next_road.sections);
									int nb_to_release_th <- length(cars_to_release);
									if length(cars_to_release)> (fs.Kmax - fs.current_concentration) {
										int nb_relase <- int (fs.Kmax - fs.current_concentration);
										cars_to_release <- nb_relase first cars_to_release;
										
										
									}
									ask cars_to_release {
										current_index <- current_index +1;
										current_target <-  last(myself.next_road.shape.points); 
									}
									myself.current_concentration <- myself.current_concentration - length(cars_to_release);
									next_road.cars <- next_road.cars + cars_to_release;
									ask fs {
										current_concentration <- current_concentration + length(cars_to_release);
									}
								}  else {
									loop rc over: cars_to_release {
										rc.active <- true;
										myself.current_concentration <- myself.current_concentration - 1;
										rc.location <- last(shape.points) ;
										rc.current_road <- nil;
										do register(agent:rc, lane:rc.current_lane);
										ask rc{do drive;} 										
									}
								}
								
							
							new_agents <- length(cars_to_release);
							if (length(cars) = new_agents) {cars <- [];}
							else {cars <- cars copy_between (new_agents, length(cars) );}
						
						}
					}
					
				} else {
					next_offer <- (self.next).offer ;
				}
				current_flow <- min([self.demand, next_offer]);
			}
			//update concentration
			ask sections   {
				float previous_flow <- (self.previous != nil) ? (self.previous).current_flow : 0.0;
				if (next != nil) {
					
					current_concentration <- current_concentration + step/myself.section_size *(previous_flow - self.current_flow) ;
				} else {
					current_concentration <- current_concentration + step/myself.section_size *(previous_flow) ;
				}
			}

	}

}

species section frequency: 0 {
	float current_concentration  ;
	float current_flow  ;
	road my_road;
	float offer ;
	float demand  ;
	bool is_last;
	rgb color;
	float remaining_car;
	float Kmax;
	float critical_concentration ; //Kc
	float max_speed ; // Vl
	float max_flow ; //Qx

	float speed_mean;
	section previous ;
	section next ;
	
	action compute_demand_offer{
		if current_concentration <= critical_concentration {
			offer <- max_flow ;
			demand <- max([0 , max_flow/critical_concentration * current_concentration ]);
		} else {
			 offer <- max([0 , - max_flow/critical_concentration * current_concentration + 2 * max_flow]);
			 demand <- max_flow ;
		}
	}
	
	aspect default {
		draw shape + 2 color: #red;
	}
	
}

	
species people skills: [advanced_driving] frequency: 0 { 
	rgb color <-rnd_color(255) ;
	node_ target_node;
	bool active <- true;
	int cpt_stucked;
	node_ current_node;
	list<node_> nodes ;
	
	
	action choose_a_path  { 
		loop while: current_path = nil {
				current_node <- one_of(node_);
				location <- current_node.location;
				target_node <-one_of(node_);
				current_path <- compute_path(graph: road_network_speed, target: target_node);
				if (current_path = nil) {
					if (current_road != nil) { 
						ask road(current_road) {
							do unregister(myself);
						}
					}
				}
			}
			if (current_path = nil) {
				do arrive;
			} 
	}
	
	
	action arrive {
		if (current_road != nil) { 
			ask road(current_road) {
				do unregister(myself);
			}
		}
		nb_arrived <- nb_arrived + 1;
		do die;
	}
	
	action driving {
		do drive;
		if ((location distance_to target_node.location) < 5.0){
			do arrive;
		}
	} 
	
	aspect base { 
		if (active) {draw triangle(20) border: #black color: color rotate:heading + 90;	}
	} 
	
} 

experiment traffic_simulation type: gui {
	output {
		monitor "nb active people" value: (people count each.active);
		monitor "nb non active people" value: (people count not each.active);
		monitor "nb arrived people" value: nb_arrived;
		monitor "total reste section:" value: sum(section collect (each.current_concentration + each.remaining_car)) ;
		display city_display2D type: opengl {
			species road aspect: carto ;
			species node_ aspect: base;
			species people aspect: base;
		}
		display charts refresh: every(10 #cycle) {
			chart "pop" type: series {
				data "nb active people" value: (people count each.active) color: #green marker: false;
				data "nb non active people" value: (people count not each.active) color: #red marker: false;
				data "nb total people" value: length(people) color: #violet marker: false;
				
				data "nb arrived people" value: nb_arrived color: #black marker: false;
				data "total reste section:" value: sum(section collect (each.current_concentration + each.remaining_car)) color: #pink marker: false ;
		
			}
		}
	}
	
}

experiment experimentation_time type: batch until: (empty(people) or time > 3#h) repeat: 10 keep_seed: true{
	parameter use_macro_factor var: use_macro_factor among: [0.0,100.0,50.0,20.0,10.0,5.0,2.0,1.0];
	parameter mode_batch var:mode_batch <- true;
	
	init {
		save ("use_macro,use_macro_factor,tot_time,end_sim") to: "results.txt" rewrite: true;
		
	}
	
	reflex info {
		float tot_time <- 0.0;
		float end_sim <- 0.0;
		loop s over: simulations {
			tot_time <- tot_time + s.time_people + s.time_roads;
			end_sim <- end_sim + s.cycle;
		}
		tot_time <- tot_time /length(simulations);
		end_sim <- end_sim /length(simulations);
		save (""+use_macro + "," + use_macro_factor +","+tot_time+ ","+end_sim) to: "results.txt" rewrite: false;
		
	}
	
}
