/**
 *  MOSAIIC Model - 9 decembre 2014
 *  Author: patricktaillandier
 *  Description: 
 */
 
model RoadTrafficComplex
 
global {   
	float critical_concentration_coeff <- 2.0;
	float max_speed <- 50 #km/#h;
	int nb_people <- 5;
	string network_type <- "leHavre" among: ["manhattan", "random", "leHavre"] ;
	bool use_macro_road <- true;
	file roads_shapefile <- network_type = "manhattan" ? file("../includes/manhattan_roads.shp") : (network_type = "random" ? file("../includes/random_roads.shp") : file("../includes/roads2.shp"));
	geometry shape <- envelope(roads_shapefile);
	graph road_network_speed;  

	float car_size <- 4.0;
	
	road first_road;
	list<node_> vertices;
	int nb_lanes <- 2 ; 
	float computation_time;
	
	float section_size <- max_speed;
	
	init {
		save "cycle,exploration_rate" to: "results/diffusion_" +network_type+ (use_macro_road ? "_macro" : "_micro") + ".csv" ;
		
		create road from: roads_shapefile with:[maxspeed::max_speed, lanes::nb_lanes] {
			create road with: [shape::line(reverse(shape.points)), maxspeed::maxspeed, lanes::lanes, linked_road::self] {
				myself.linked_road <- self;
			}
		}
		first_road <- road with_min_of ({0,0} distance_to first(each.shape.points));
		first_road.visited <- true;
		list<point> points <- [];
		ask road {
			point fp <- first(shape.points);
			point lp <- last(shape.points);
			if not(fp in points) {create node_ with:[location::fp];points << fp;}
			if not(lp in points) {create node_ with:[location::lp];points << lp;}
			
		}
		road_network_speed <-  (as_driving_graph(road, node_)) ;
		ask road {
			next_roads <- list<road> (node_(self.target_node).roads_out);
			next_roads >> self.linked_road;	
			prev_roads <- list<road> (node_(self.source_node).roads_in);
		}
		if (use_macro_road) {
			do create_sections;
		}
	}
	reflex add_drivers when: time = 0 and every(5 #mn) {
		if (not use_macro_road) {
			loop times: nb_people {
				do create_people(first_road, rnd(nb_lanes));
			}
		} else {
			ask first(first_road.sections) {
				current_concentration <- current_concentration + nb_people;
			}
		}
	}
	list<float> exploration;
	reflex record_explo when: every(1 #mn) {
		exploration << ((road count each.visited) / length(road));
	}
	/*reflex end when: time = 1 #h {
		write "diffusion_" +network_type+ (use_macro_road ? "_macro" : "_micro") + " time: " + computation_time;
		do pause;
	}*/
	action create_people (road the_road, int lane) {
		create people  { 
			speed <- max_speed ;
			real_speed <- 130 #km /#h ;
			vehicle_length <- car_size;
			right_side_driving <- true;
			proba_lane_change_up <-rnd(0.5,1.0);
			proba_lane_change_down <- rnd(0.5,1.0);
			current_node <- first(node_);
			location <- first(the_road.shape.points);
			current_lane <- lane;
			security_distance_coeff <- rnd(1.5,3.0);  
			proba_respect_priorities <- 0.5;
			proba_respect_stops <- [1.0];
			proba_block_node <- 0.0;
			proba_use_linked_road <- 0.0;
			max_acceleration <- rnd(10.0,15.0) #km/#h;
			max_speed <- 150 #km /#h ;
			speed_coeff <- rnd(0.8,1.1);
			ask the_road {
				do register(myself,lane);
			}
		}
	}
	
	action create_sections {
		ask road {
			is_macro <- true;
			int nb <- max([1,round (shape.perimeter /section_size)]);
			float sec_length <- shape.perimeter / nb;
			section prev <- nil;
			int cpt <- 0;
			list<point> pts <- points_on (shape,sec_length);
			list<geometry> geoms;
			geometry g <- copy(shape);
			loop pt over: pts {
				if (pt distance_to first(shape.points) > (section_size/10.0)) and (pt distance_to last(shape.points) > (section_size/10.0)) {
					list<geometry> gg <- g split_at pt;
					geoms<< gg[0];
					g <- gg[1];
				}
			}
			geoms << g;
			loop times: nb {
				create section with:[location::location]{
					max_speed <- myself.maxspeed #km/#h;
					color <-  #red;
					previous <- prev;
					if (previous != nil) {previous.next <- self;}
					prev <- self;
					Kmax <- sec_length /  (car_size + 1.0) ;
					critical_concentration <- Kmax / critical_concentration_coeff; //Kc
					current_concentration <- 0.0 ;
					current_flow <- current_concentration * max_speed ;
					max_flow <- critical_concentration * max_speed ; 
					myself.sections << self;
					my_road <- myself;
					point pt <- first(node_).location;
					shape <- geoms[cpt];
					cpt <- cpt + 1;
					
				}	
			}
			regular_sections <- sections where ((each.next != nil) and (each.previous != nil));
			ask target_node as: node_{
				inputSections << last(myself.sections);
			}
			ask source_node as: node_{
				outputSections << first(myself.sections);
			}
		} 
	}
	
	reflex manage_macro_raods when: use_macro_road{
		float t<-machine_time;
		ask road {concentration_tot <- sections sum_of each.current_concentration;}
		list<road> active_roads <- road where (each.is_active);
		ask active_roads {visited <- true;}
		list<section> active_sections <- active_roads accumulate each.sections;
		
		ask active_sections {
			do compute_demand_offer;
			}
		
		ask node_ {
			do manage_flow;
		}
		
		ask active_roads {
			do manage_flow;
		}
		
		ask node_ {
			do manage_flow_end;
		}
		computation_time <- computation_time + machine_time - t;
	}
} 


	
species node_ skills: [skill_road_node] {
	list<section> inputSections;
	list<section> outputSections;
	aspect base {    
		draw square(4) color: #magenta;
	} 
	
	action manage_flow {
		float sum_demand <- sum(inputSections collect each.demand);
		float remaining_demand <- copy(sum_demand);
		ask outputSections where (each.next != nil){
			next_offer <- (self.next).offer ;
			current_flow <- min([self.demand, next_offer]);
			offer_tmp <- 0.0;
		}
		list<section> possible_sections <- outputSections where (each.offer > each.offer_tmp);
		loop while: (remaining_demand > 0) and (not empty(possible_sections)) {
			possible_sections <- outputSections where (each.offer > each.offer_tmp);
			float to_demand <- not empty(possible_sections) ? (remaining_demand / length(possible_sections)) with_precision 5 : 0.0;
			if (to_demand = 0) {break;}
			ask possible_sections {
				float to_give <- min([to_demand, offer - offer_tmp]);
				offer_tmp <- offer_tmp + to_give;
				remaining_demand <- remaining_demand - to_give;
			}	
		}
		ask outputSections {
			current_concentration <- current_concentration + step/section_size *(offer_tmp - self.current_flow) ;
		}
		
		ask inputSections {
			current_flow <- 0.0;
		}
		possible_sections <- inputSections where (each.demand > each.current_flow);
		float real_demand <- sum_demand - remaining_demand;
		loop while: (real_demand > 0) and (not empty(possible_sections)) {
			possible_sections <- inputSections where (each.demand > each.current_flow);
			float to_demand <- not empty(possible_sections) ? (real_demand / length(possible_sections))with_precision 5: 0.0;
			if (to_demand = 0) {break;}
			ask possible_sections {
				float to_give <- min([to_demand, demand - current_flow]);
				current_flow <- current_flow + to_give;
				real_demand <- real_demand - to_give;
			}	
		}
		
				
	}
	
	action manage_flow_end{
		ask inputSections where (each.previous != nil){
			float previous_flow <- (self.previous).current_flow;
			current_concentration <- current_concentration + step/section_size *(previous_flow - self.current_flow) ;
		}
	}
}

species road skills: [skill_road] frequency: 0{
	string oneway;
	int nb_people <- 0;
	int nb_people_tot <- 0;
	float concentration_tot;
	
	string highway;
	list<node_> neighbours;
	
	list<section> sections;
	bool is_macro <- false;
	float sum_concentration <-0.0;
			
	list<people> cars_to_capture <- [];
	list<road> next_roads;
	list<section> regular_sections;
	list<road> prev_roads;
	bool is_active -> {concentration_tot > 0 or ((prev_roads first_with (each.concentration_tot > 0)) != nil)};
	bool visited <- false;
	aspect carto {
		if not(visited) and not(linked_road.visited)  {
			draw shape color: #black;
		} else {
			draw shape color: #red;
		}
	}
	
	
	action manage_flow  {
		ask regular_sections{ 
			next_offer <- (self.next).offer ;
			current_flow <- min([self.demand, next_offer]);
		}
		ask regular_sections {
			float previous_flow <- (self.previous).current_flow;
			current_concentration <- current_concentration + step/section_size *(previous_flow - self.current_flow) ;
		}
	}
	
}

species section schedules:[]{
	float current_concentration  ;
	float current_flow  ;
	road my_road;
	float offer ;
	float demand  ;
	bool is_last;
	rgb color;
	float remaining_car;
	float Kmax;
	float critical_concentration;
	float max_speed ; // Vl
	float max_flow ; //Qx
	float next_offer;
	float offer_tmp;

	float speed_mean;
	section previous ;
	section next ;
	bool is_regular <- false;
	geometry shape;
	
	action compute_demand_offer{
		next_offer <- 0.0;
		if current_concentration <= critical_concentration {
			offer <- max_flow ;
			demand <- max([0 , max_flow/critical_concentration * current_concentration ]);
		} else {
			 offer <- max([0 , - max_flow/critical_concentration * current_concentration + 2 * max_flow]);
			 demand <- max_flow ;
		}
	}
}

	
species people skills: [advanced_driving] schedules: (people sort_by (- 1000000 * each.segment_index_on_road + each.distance_to_goal)) {
	rgb color <-rnd_color(255) ;
	node_ target_node;
	
	bool is_stopped <- false;
	bool recompute_path <- false;
	bool to_delete <- false;
	float proba_avoid_traffic_jam;
	node_ current_node;
	bool mode_avoid <- false;
	int cpt_avoid <- 0;
		
	reflex driving {
		float t <- machine_time;
		do drive_random;
		computation_time <- computation_time + machine_time - t;
		road(current_road).visited <- true;
	} 
	
	aspect base { 
		draw triangle(40) color: color rotate:heading + 90;	
	} 
	
	
} 

experiment traffic_simulation type: gui {
	output {
		
		display city_display2D type: opengl scale: true draw_env: false{
			graphics "first_road" {
				if (first_road != nil) {
					draw first_road.shape + 50.0 color: #red;
				}
			}
			species road aspect: carto ;
			
			species node_ aspect: base;
			species people aspect: base; 
		}
	}
	
}

experiment experimentation_time type: batch until: (time > 1#h) repeat: 10 keep_seed: true{
	parameter network_type  var: network_type among: ["manhattan", "random", "leHavre"] ;
	parameter use_macro_road var: use_macro_road among: [true, false];
	
	init {
		save "network_type,use_macro_road,tot_time" to: "diffusion_time.csv" rewrite: true;
		
	}
	
	reflex info {
		float tot_time <- 0.0;
		list<float> explo <- [];
		loop times: length(simulations[0].exploration) {explo<<0.0;}
		loop s over: simulations {
			tot_time <- tot_time + s.computation_time;
			loop i from: 0 to: length(s.exploration) -1 {explo[i] <- explo[i] + s.exploration[i];}
		}
		loop i from: 0 to: length(explo) -1 {explo[i] <- explo[i] / length(simulations);}
		
		tot_time <- tot_time /length(simulations);
		loop i from: 0 to: length(explo) -1 {
			save ""+ (i*60)+"," + explo[i] to: "diffusion_" +network_type+ (use_macro_road ? "_macro" : "_micro") + ".csv" rewrite: false;
		}
		
		save ""+ network_type+"," + use_macro_road + "," + tot_time to: "diffusion_time.csv" rewrite: false;
	
		
	}
	
}

