/**
 *  MOSAIIC Model - 9 decembre 2014
 *  Author: patricktaillandier
 *  Description: 
 */
 
model RoadTrafficComplex
 
global {   
	float critical_concentration_coeff <- 2.0;
	float max_speed <- 90 #km/#h;
	geometry shape <- rectangle(size_road + 20.0,size_road +  20);
	graph road_network_speed;  
	int nb_steps_stats <- 60;
	
	bool record <- false ;
	int cpt_record;
	int cpt_record_max <- 500;
	int cpt_change_nb update: cpt_change_nb + 1;
	
	float car_size <- 4.0;
	
	
	float min_length <- 0.0;
	
 	list<node_> vertices;
	float size_road <- max_speed * 10 ;
	int nb_lanes <- 2 ; 
	
	float section_size <- max_speed ;//100#m ; // DeltaX
	bool use_macro_road <- true parameter: true;
	
	
	float nb_people_coeff <- 0.05;
	int nb_people_tot;
	bool first <- true;
		
	init {
		do initialization;
	}
	
	int cycle_tot_ag<-0;
	int cycle_tot_macro<-0;
	
	float compuation_time;
	
	action initialization {
		//save "concentration,speed,flow" to: "results/circuit_" + (use_macro_road ? "macro" : "micro") + ".csv" ;
		//save "nb_drivers,compuation_time" to: "results/circuit_compuation_" + (use_macro_road ? "macro" : "micro") + ".csv";
		
		create node_ with:[location:: {10,size_road/2.0+10}, is_traffic_signal::false, is_crossing :: false];
		create node_ with:[location:: {size_road+10,size_road/2.0+10}, is_traffic_signal::false, is_crossing :: false];
		create road with:[shape:: line([first(node_).location, first(node_).location - {0.0,size_road/2.0}, (node_[1]).location-{0.0,size_road/2.0},(node_[1]).location]),highway::"motorway",lanes::nb_lanes, maxspeed::max_speed] {
			capacite_max <- 1+ int(lanes * shape.perimeter/(4.0));
			
		}
		create road with:[shape:: line([(node_[1]).location,(node_[1]).location + {0.0,size_road/2.0}, first(node_).location+{0.0,size_road/2.0}, first(node_).location]),highway::"motorway",lanes::nb_lanes, maxspeed::max_speed] {
			capacite_max <- 1+ int(lanes * shape.perimeter/(4.0));
			
		}
		map general_speed_map_speed <- road as_map (each::(each.shape.perimeter / each.maxspeed)); 
		
		
		road_network_speed <-  (as_driving_graph(road, node_))  with_weights general_speed_map_speed;
		
	
		list<string> roads_importance <- ["service","residential", "unclassified",  "tertiary_link", "tertiary",  "secondary_link", "secondary", "primary_link", "primary", "trunk_link", "trunk", "motorway_link", "motorway"];
		do create_sections;
		
		
		int section_id <- round(length(section)/4.0);
		create counting_loop {
			name <- "section "+ section_id;
			my_section <- empty(section) ? nil : section[section_id];
			shape <- my_section.shape;
			distance_to_goal_start <- first(shape.points) distance_to last(node_);
			distance_to_goal_end <- last(shape.points) distance_to last(node_);
			color <- rnd_color(255);
	
		}
	}
	
	reflex to_record when: not record{
		cpt_record <- cpt_record + 1;
		if (cpt_record = cpt_record_max){
			cpt_record <- 0;
			record <- true;
			cpt_change_nb <- 0;
		} 
	}
	
	
	
	reflex create_people when: record and (cpt_change_nb = (nb_steps_stats * 10))  {
		if (nb_people_tot = sum(section collect each.Kmax)) {
			//do pause;
		} else {
			first <- true;
			int nb_people <- int(nb_people_coeff * sum(section collect each.Kmax));
			nb_people_tot <- nb_people_tot + nb_people;
			int entier <- int(nb_people/ length(section));
			int reste <- nb_people - entier;
			loop sect over: section {
				loop times: entier{
					if (not use_macro_road) {
						do create_people(sect, rnd(nb_lanes));
					} else {
						ask sect {
							current_concentration <- current_concentration + 1;
						}
					}
				}
			}
			loop times: reste{
				section sect <- one_of(section);
				if (not use_macro_road) {
					do create_people(sect, rnd(nb_lanes));
				} else {
					ask sect {
						current_concentration <- current_concentration + 1;
					}
				}
			}
			record <- false;
		}
		
	}
	
	
	
	action create_people (section sect, int lane) {
		create people  { 
			speed <- max_speed ;
			real_speed <- 130 #km /#h ;
			vehicle_length <- car_size;
			right_side_driving <- true;
			proba_lane_change_up <-rnd(0.5,1.0);
			proba_lane_change_down <- rnd(0.5,1.0);
			current_node <- first(node_);
			location <- any_location_in(sect);
			target_node <- last(node_);
			distance_to_goal <- location distance_to target_node;
			current_lane <- lane;
			security_distance_coeff <- rnd(1.5,3.0);  
			proba_respect_priorities <- 1.0;
			proba_respect_stops <- [1.0];
			proba_block_node <- 0.0;
			proba_use_linked_road <- 0.0;
			max_acceleration <- rnd(10.0,15.0) #km/#h;
			max_speed <- 150 #km /#h ;
			speed_coeff <- rnd(0.8,1.1);
			road rd <- sect.my_road;
			ask rd {
				do register(myself,0);
			}
		}
	}
	
	action create_sections {
		ask road where (each.shape.perimeter >= section_size ) {
			is_macro <- true;
			int nb <- round (shape.perimeter /section_size);
			float sec_length <- shape.perimeter / nb;
			section prev <- nil;
			int cpt <- 0;
			list<point> pts <- points_on (shape,sec_length);
			list<geometry> geoms;
			geometry g <- copy(shape);
			loop pt over: pts {
				if (pt distance_to first(shape.points) > (section_size/10.0)) and (pt distance_to last(shape.points) > (section_size/10.0)) {
					list<geometry> gg <- g split_at pt;
					geoms<< gg[0];
					g <- gg[1];
				}
			}
			geoms << g;
			loop times: nb {
				create section with:[location::location]{
					max_speed <- myself.maxspeed #km/#h;
					color <-  #red;
					previous <- prev;
					if (previous != nil) {previous.next <- self;}
					prev <- self;
					Kmax <- myself.lanes * sec_length /  (car_size) ;
					critical_concentration <- Kmax / critical_concentration_coeff; //Kc
					current_concentration <- 0.0 ;
					current_flow <- current_concentration * max_speed ;
					max_flow <- critical_concentration * max_speed ; 
					myself.sections << self;
					my_road <- myself;
					point pt <- first(node_).location;
					shape <- geoms[cpt];
					cpt <- cpt + 1;
					
				}	
			}
			regular_sections <- sections where ((each.next != nil) and (each.previous != nil));
			ask target_node as: node_{
				inputSections << last(myself.sections);
			}
			ask source_node as: node_{
				outputSections << first(myself.sections);
			}
		} 
	}
	
	
	reflex manage_macro_roads when: use_macro_road{
		float t <- machine_time;
		ask section {
			previous_concentration <- current_concentration;
			do compute_demand_offer;	
		}
		
		ask node_ {
			do manage_flow;
		}
		
		ask road {
			do manage_flow;
		}
		
		ask node_ {
			do manage_flow_end;
		}
		ask section {
			speed_mean <- previous_concentration = 0 ? 0.0 : current_flow/previous_concentration ;
		}
		compuation_time <- compuation_time + machine_time - t;
		
	}
} 



species node_ skills: [skill_road_node] {
	list<section> inputSections;
	list<section> outputSections;
	aspect base {    
		draw square(4) color: #magenta;
	} 
	
	action manage_flow {
		float sum_demand <- sum(inputSections collect each.demand);
		float remaining_demand <- copy(sum_demand);
		ask outputSections {
			next_offer <- (self.next).offer ;
			current_flow <- min([self.demand, next_offer]);
			offer_tmp <- 0.0;
		}
		list<section> possible_sections <- outputSections where (each.offer > each.offer_tmp);
		loop while: (remaining_demand > 0) and (not empty(possible_sections)) {
			possible_sections <- outputSections where (each.offer > each.offer_tmp);
			float to_demand <- not empty(possible_sections) ? (remaining_demand / length(possible_sections)) with_precision 2 : 0.0;
			if (to_demand = 0) {break;}
			ask possible_sections {
				float to_give <- min([to_demand, offer - offer_tmp]);
				offer_tmp <- offer_tmp + to_give;
				remaining_demand <- remaining_demand - to_give;
			}	
		}
		ask outputSections {
			current_concentration <- current_concentration + step/section_size *(offer_tmp - self.current_flow) ;
		}
		
		ask inputSections {
			current_flow <- 0.0;
		}
		possible_sections <- inputSections where (each.demand > each.current_flow);
		float real_demand <- sum_demand - remaining_demand;
		loop while: (real_demand > 0) and (not empty(possible_sections)) {
			possible_sections <- inputSections where (each.demand > each.current_flow);
			float to_demand <- not empty(possible_sections) ? (real_demand / length(possible_sections))with_precision 2 : 0.0;
			if (to_demand = 0) {break;}
			ask possible_sections {
				float to_give <- min([to_demand, demand - current_flow]);
				current_flow <- current_flow + to_give;
				real_demand <- real_demand - to_give;
			}	
		}
		
				
	}
	
	action manage_flow_end{
		ask inputSections {
			float previous_flow <- (self.previous).current_flow;
			current_concentration <- current_concentration + step/section_size *(previous_flow - self.current_flow) ;
		}
	}
}

species road skills: [skill_road] {
	string oneway;
	int nb_people <- 0;
	int nb_people_tot <- 0;
	
	int capacite_max ;
	float nb_bloques <- 0.0;
	int min_traffic_jam_people_destroy;
	int min_traffic_jam_people_creation;
	
	road next_linked_road;
	string highway;
	list<node_> neighbours;
	
	list<section> sections;
	bool is_macro <- false;
			
	list<people> cars_to_capture <- [];
	list<road> next_roads;
	list<section> regular_sections;
	
	aspect carto {
		if not(use_macro_road) {
			draw shape color: #black;
		} else {
			loop sect over: sections {
				draw sect.shape + (1 + 3 * sect.current_concentration/sect.critical_concentration ) border: #black color: #red;
			}
		}
	}
	
	
	action manage_flow  {
		ask regular_sections{ 
			next_offer <- (self.next).offer ;
			current_flow <- min([self.demand, next_offer]);
		}
		ask regular_sections {
			float previous_flow <- (self.previous).current_flow;
			current_concentration <- current_concentration + step/section_size *(previous_flow - self.current_flow) ;
			
		}
	}
	
}

species section schedules:[]{
	float current_concentration  ;
	float current_flow  ;
	float previous_concentration;
	road my_road;
	float offer ;
	float demand  ;
	bool is_last;
	rgb color;
	float remaining_car;
	float Kmax;
	float critical_concentration;
	float max_speed ; // Vl
	float max_flow ; //Qx
	float next_offer;
	float offer_tmp;

	float speed_mean;
	section previous ;
	section next ;
	
	geometry shape;
	
	action compute_demand_offer{
		next_offer <- 0.0;
		if current_concentration <= critical_concentration {
			offer <- max_flow ;
			demand <- max([0 , max_flow/critical_concentration * current_concentration ]);
		} else {
			 offer <- max([0 , - max_flow/critical_concentration * current_concentration + 2 * max_flow]);
			 demand <- max_flow ;
		}
	}
}

	
species people skills: [advanced_driving] schedules: (people sort_by (- 1000000 * each.segment_index_on_road + each.distance_to_goal)) {
	rgb color <-rnd_color(255) ;
	node_ target_node;
	
	bool is_stopped <- false;
	bool recompute_path <- false;
	bool to_delete <- false;
	float proba_avoid_traffic_jam;
	node_ current_node;
	bool mode_avoid <- false;
	int cpt_avoid <- 0;
		
	reflex driving {
		float t <- machine_time;
		do drive_random;
		compuation_time <- compuation_time + machine_time - t;
	} 
	
	aspect base { 
		draw triangle(4) color: color rotate:heading + 90;	
	} 
	
	
	float external_factor_impact(road new_road,float remaining_time) {
		if (current_path != nil) {
			new_road.nb_people_tot <- new_road.nb_people_tot + 1;
		}
		return remaining_time;
	}
	
} 

species counting_loop {
	float distance_to_goal_start;
	float distance_to_goal_end;
	list<float> numbers_in;
	list<float> numbers_out;
	list<float> numbers;
	list<float> mean_speeds;
	list<float> numbers_in_step;
	list<float> numbers_out_step;
	list<float> numbers_step;
	list<float> mean_speeds_step;
	list<float> mean_flows_step;
	section my_section;
	float number_in;
	float number_out;
	float number;
	float mean_speed;
	float mean_flow;
	rgb color;
	list<people> people_on_previous;
	aspect default {
		draw shape + 2 color: color;
	}
	reflex count_micro when: record and not use_macro_road{
		list<people> people_on <- people where ((each.distance_to_goal <= distance_to_goal_start) and (each.distance_to_goal >= distance_to_goal_end));
		numbers_in_step << float(people_on count !(each in people_on_previous));
		int nb_out <- people_on_previous count !(each in people_on);
		numbers_out_step << float(nb_out);
		mean_flows_step << (nb_out/step);
		people_on_previous <- copy(people_on);
		numbers_step << float(length(people_on));
		if (not empty(people_on)) {mean_speeds_step << mean(people_on collect each.real_speed)  #h/#km;}
		
		
	}
	
	reflex count_macro when: record and use_macro_road{
		numbers_in_step << step/section_size * my_section.previous.current_flow;
		numbers_out_step << step/section_size * my_section.current_flow;
		numbers_step << my_section.current_concentration;
		mean_flows_step << my_section.current_flow / section_size;
		if (my_section.speed_mean > 0) {mean_speeds_step << my_section.speed_mean * #h /#km * #h /#km ;}
	}
	
	reflex stats when: record and (cpt_change_nb mod (nb_steps_stats ) = 0)  and cycle > 0{
		number_in<-mean(numbers_in_step);
		number_out<-mean(numbers_out_step);
		number<-mean(numbers_step);
		mean_speed<-mean(mean_speeds_step);
		mean_flow <- mean(mean_flows_step);
		
		numbers_in<<number_in;
		numbers_out<<number_out;
		numbers<<number;
		mean_speeds<<mean_speed;
			
		numbers_in_step <- [];
		numbers_out_step<- [];
		numbers_step<- [];
		mean_speeds_step<- [];
		mean_flows_step <- [];
		if (not first and nb_people_tot > 0) {
			save (""+number+"," +mean_speed + ","+ mean_flow) to: "circuit_" + (use_macro_road ? "macro" : "micro") + ".csv" rewrite: false;
			save "" + nb_people_tot + "," + compuation_time to: "circuit_compuation_" + (use_macro_road ? "macro" : "micro") + ".csv" rewrite: false;
		}
		first <- false;
		compuation_time <- 0.0;
		
	}
	
}

experiment traffic_simulation type: gui {
	output {
		
		display city_display2D {
			species counting_loop refresh: false;
			species road aspect: carto ;
			species node_ aspect: base;
			species people aspect: base; 
			
		}
		
		display charts refresh: record and (cpt_change_nb mod (nb_steps_stats ) = 0) {
			chart "counting loop number_in" type: series size:{0.5,0.5} position:{0.0,0.0} {
				loop cl over: counting_loop {
					data cl.name value: (cl.number_in) color: cl.color;
				}
			}
			chart "counting loop number out" type: series size:{0.5,0.5} position:{0.5,0.0} {
				loop cl over: counting_loop {
					data cl.name value: (cl.number_out) color: cl.color;
				}
			}
			chart "counting loop number" type: series size:{0.5,0.5} position:{0.0,0.5} {
				loop cl over: counting_loop {
					data cl.name value: (cl.number) color: cl.color;
				}
			}
			chart "counting loop mean speed" type: series size:{0.5,0.5} position:{0.5,0.5} {
				loop cl over: counting_loop {
					data cl.name value: (cl.mean_speed) color: cl.color;
				}
			}
		}
		
		display phase_diagram refresh: record and (cpt_change_nb mod (nb_steps_stats ) = 0)  {
			chart "concentration_speed" type: xy  size:{0.5,0.5} position:{0.0,0.0} x_label: 'Concentration (nb cars)' y_label: 'Vitesse (km/h)'
			{
				data legend: "concentration_speed" value: [first(counting_loop).number,first(counting_loop).mean_speed,1]  color: #green style: dot;
			}
			chart "concentration_flow" type: xy  size:{0.5,0.5} position:{0.5,0.0} x_label: 'Concentration (nb cars)' y_label: 'Debit (nb cars/s)'
			{
				data legend: "concentration_flow" value: [first(counting_loop).number,first(counting_loop).mean_flow,1] color: #red style: dot;
			}
			chart "speed_flow" type: xy  size:{0.5,0.5} position:{0.0,0.5} x_label: 'Speed (km/h)' y_label: 'Debit (nb cars/s)'
			{
				data legend: "speed_flow" value: [first(counting_loop).mean_speed,first(counting_loop).mean_flow,1] color: #blue style: dot;
			}
		}
	}
	
}


experiment experimentation_time type: batch until: record and (cpt_change_nb = (nb_steps_stats * 10))  and (nb_people_tot = sum(section collect each.Kmax))  repeat: 10 keep_seed: true{
	parameter use_macro_road var: use_macro_road among: [true, false];
	
	init {
		save "concentration,speed,flow" to: "circuit_" + (use_macro_road ? "macro" : "micro") + ".csv" ;
		save "nb_drivers,compuation_time" to: "circuit_compuation_" + (use_macro_road ? "macro" : "micro") + ".csv";
	
		
	}
	
}



